// <file>
//     <copyright see="prj:///doc/copyright.txt"/>
//     <license see="prj:///doc/license.txt"/>
//     <owner name="Mike Krüger" email="mike@icsharpcode.net"/>
//     <version value="$version"/>
// </file>

using System;
using System.Collections;

namespace MonoDevelop.Projects.Parser
{
	/// <summary>
	///     <para>
	///       A collection that stores <see cref='.ClassProxy'/> objects.
	///    </para>
	/// </summary>
	/// <seealso cref='.ClassProxyCollection'/>
	[Serializable()]
	internal class ClassProxyCollection : CollectionBase 
	{
		Hashtable nameHashtable                = new Hashtable();
		Hashtable caseInsensitiveNameHashtable = new Hashtable();
		/// <summary>
		///     <para>
		///       Initializes a new instance of <see cref='.ClassProxyCollection'/>.
		///    </para>
		/// </summary>
		public ClassProxyCollection() {
		}
		
		/// <summary>
		///     <para>
		///       Initializes a new instance of <see cref='.ClassProxyCollection'/> based on another <see cref='.ClassProxyCollection'/>.
		///    </para>
		/// </summary>
		/// <param name='value'>
		///       A <see cref='.ClassProxyCollection'/> from which the contents are copied
		/// </param>
		public ClassProxyCollection(ClassProxyCollection value) {
			this.AddRange(value);
		}
		
		/// <summary>
		///     <para>
		///       Initializes a new instance of <see cref='.ClassProxyCollection'/> containing any array of <see cref='.ClassProxy'/> objects.
		///    </para>
		/// </summary>
		/// <param name='value'>
		///       A array of <see cref='.ClassProxy'/> objects with which to intialize the collection
		/// </param>
		public ClassProxyCollection(ClassProxy[] value) {
			this.AddRange(value);
		}
		
		/// <summary>
		/// <para>Represents the entry at the specified index of the <see cref='.ClassProxy'/>.</para>
		/// </summary>
		/// <param name='index'><para>The zero-based index of the entry to locate in the collection.</para></param>
		/// <value>
		///    <para> The entry at the specified index of the collection.</para>
		/// </value>
		/// <exception cref='System.ArgumentOutOfRangeException'><paramref name='index'/> is outside the valid range of indexes for the collection.</exception>
		public ClassProxy this[int index] {
			get {
				return (ClassProxy)List[index];
			}
			set {
				List[index] = value;
			}
		}
		
		/// <summary>
		///    <para>Adds a <see cref='.ClassProxy'/> with the specified value to the
		///    <see cref='.ClassProxyCollection'/> .</para>
		/// </summary>
		/// <param name='value'>The <see cref='.ClassProxy'/> to add.</param>
		/// <returns>
		///    <para>The index at which the new element was inserted.</para>
		/// </returns>
		/// <seealso cref='.ClassProxyCollection.AddRange'/>
		public int Add(ClassProxy value) 
		{
			nameHashtable[value.FullyQualifiedName.GetHashCode()]                          = Count;
			caseInsensitiveNameHashtable[value.FullyQualifiedName.ToLower().GetHashCode()] = Count;
			
			return List.Add(value);
		}
		
		/// <summary>
		/// <para>Copies the elements of an array to the end of the <see cref='.ClassProxyCollection'/>.</para>
		/// </summary>
		/// <param name='value'>
		///    An array of type <see cref='.ClassProxy'/> containing the objects to add to the collection.
		/// </param>
		/// <returns>
		///   <para>None.</para>
		/// </returns>
		/// <seealso cref='.ClassProxyCollection.Add'/>
		public void AddRange(ClassProxy[] value) 
		{
			for (int i = 0; (i < value.Length); i = (i + 1)) {
				this.Add(value[i]);
			}
		}
		
		/// <summary>
		///     <para>
		///       Adds the contents of another <see cref='.ClassProxyCollection'/> to the end of the collection.
		///    </para>
		/// </summary>
		/// <param name='value'>
		///    A <see cref='.ClassProxyCollection'/> containing the objects to add to the collection.
		/// </param>
		/// <returns>
		///   <para>None.</para>
		/// </returns>
		/// <seealso cref='.ClassProxyCollection.Add'/>
		public void AddRange(ClassProxyCollection value) {
			for (int i = 0; (i < value.Count); i = (i + 1)) {
				this.Add(value[i]);
			}
		}
		
		/// <summary>
		/// <para>Gets a value indicating whether the
		///    <see cref='.ClassProxyCollection'/> contains the specified <see cref='.ClassProxy'/>.</para>
		/// </summary>
		/// <param name='value'>The <see cref='.ClassProxy'/> to locate.</param>
		/// <returns>
		/// <para><see langword='true'/> if the <see cref='.ClassProxy'/> is contained in the collection;
		///   otherwise, <see langword='false'/>.</para>
		/// </returns>
		/// <seealso cref='.ClassProxyCollection.IndexOf'/>
		public bool Contains(ClassProxy value) 
		{
			return List.Contains(value);
		}
		
		/// <summary>
		/// <para>Copies the <see cref='.ClassProxyCollection'/> values to a one-dimensional <see cref='System.Array'/> instance at the
		///    specified index.</para>
		/// </summary>
		/// <param name='array'><para>The one-dimensional <see cref='System.Array'/> that is the destination of the values copied from <see cref='.ClassProxyCollection'/> .</para></param>
		/// <param name='index'>The index in <paramref name='array'/> where copying begins.</param>
		/// <returns>
		///   <para>None.</para>
		/// </returns>
		/// <exception cref='System.ArgumentException'><para><paramref name='array'/> is multidimensional.</para> <para>-or-</para> <para>The number of elements in the <see cref='.ClassProxyCollection'/> is greater than the available space between <paramref name='arrayIndex'/> and the end of <paramref name='array'/>.</para></exception>
		/// <exception cref='System.ArgumentNullException'><paramref name='array'/> is <see langword='null'/>. </exception>
		/// <exception cref='System.ArgumentOutOfRangeException'><paramref name='arrayIndex'/> is less than <paramref name='array'/>'s lowbound. </exception>
		/// <seealso cref='System.Array'/>
		public void CopyTo(ClassProxy[] array, int index) {
			List.CopyTo(array, index);
		}
		
		/// <summary>
		///    <para>Returns the index of a <see cref='.ClassProxy'/> in
		///       the <see cref='.ClassProxyCollection'/> .</para>
		/// </summary>
		/// <param name='value'>The <see cref='.ClassProxy'/> to locate.</param>
		/// <returns>
		/// <para>The index of the <see cref='.ClassProxy'/> of <paramref name='value'/> in the
		/// <see cref='.ClassProxyCollection'/>, if found; otherwise, -1.</para>
		/// </returns>
		/// <seealso cref='.ClassProxyCollection.Contains'/>
		public int IndexOf(ClassProxy value) 
		{
			return List.IndexOf(value);
		}
		
		public int IndexOf(string fullyQualifiedName, bool caseSensitive)
		{
			object o = caseSensitive ? nameHashtable[fullyQualifiedName.GetHashCode()] : caseInsensitiveNameHashtable[fullyQualifiedName.ToLower().GetHashCode()];
			if (o != null) {
				return (int)o;
			}
			return -1;
		}
		
		/// <summary>
		///    <para>Returns an enumerator that can iterate through
		///       the <see cref='.ClassProxyCollection'/> .</para>
		/// </summary>
		/// <returns><para>None.</para></returns>
		/// <seealso cref='System.Collections.IEnumerator'/>
		public new ClassProxyEnumerator GetEnumerator() 
		{
			return new ClassProxyEnumerator(this);
		}
		
		/// <summary>
		///    <para> Removes a specific <see cref='.ClassProxy'/> from the
		///    <see cref='.ClassProxyCollection'/> .</para>
		/// </summary>
		/// <param name='value'>The <see cref='.ClassProxy'/> to remove from the <see cref='.ClassProxyCollection'/> .</param>
		/// <returns><para>None.</para></returns>
		/// <exception cref='System.ArgumentException'><paramref name='value'/> is not found in the Collection. </exception>
		public void Remove(ClassProxy value) 
		{
			List.Remove(value);
		}
		
		public class ClassProxyEnumerator : object, IEnumerator {
			
			private IEnumerator baseEnumerator;
			
			private IEnumerable temp;
			
			public ClassProxyEnumerator(ClassProxyCollection mappings) {
				this.temp = ((IEnumerable)(mappings));
				this.baseEnumerator = temp.GetEnumerator();
			}
			
			public ClassProxy Current {
				get {
					return ((ClassProxy)(baseEnumerator.Current));
				}
			}
			
			object IEnumerator.Current {
				get {
					return baseEnumerator.Current;
				}
			}
			
			public bool MoveNext() {
				return baseEnumerator.MoveNext();
			}
			
			bool IEnumerator.MoveNext() {
				return baseEnumerator.MoveNext();
			}
			
			public void Reset() {
				baseEnumerator.Reset();
			}
			
			void IEnumerator.Reset() {
				baseEnumerator.Reset();
			}
		}
	}
}
